# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Model Utils Class."""

import os
import subprocess
import shutil
import stat
import sys
import tempfile
from collections import namedtuple
from pathlib import Path
from typing import Callable, Dict, List
from azureml.assets.util import logger
from azure.ai.ml._azure_environments import (
    AzureEnvironments,
    _get_default_cloud_name,
    _get_storage_endpoint_from_metadata
)

ReplacePair = namedtuple('ReplacePair', ['Original', 'Replacement'])


class CopyUpdater:
    """Update files during azcopy."""

    def __init__(self):
        """Initialize copy updater."""
        self._file_functions: Dict[str, Callable[[Path], bool]] = {}

    def add_file_function(self, file_name: str, func: Callable[[Path], bool]):
        """Add a file function.

        Args:
            file_name (str): Path to file to update, relative to the source path, using POSIX directory separators.
            func (Callable[[Path], bool]): Function that will receive the path to the file to update and returns True
                                           if an update was made, False otherwise.
        """
        self._file_functions[file_name] = func

    @property
    def files(self) -> List[str]:
        """Return the list of files to update."""
        return list(self._file_functions.keys())

    def update_files(self, root_path: Path) -> bool:
        """Update files under the given path.

        Args:
            root_path (Path): Path containing files to update.

        Returns:
            bool: True if any file was updated, False otherwise.
        """
        updated = False
        for file_name, func in self._file_functions.items():
            file_path = root_path / file_name
            if file_path.exists():
                logger.print(f"Updating {file_path}")
                updated |= func(file_path)
            else:
                logger.log_warning(f"{file_path} not found")
        return updated

    @staticmethod
    def create_replace_function(*pairs: ReplacePair) -> Callable[[Path], bool]:
        """Create a function to replace text in a file.

        Args:
            pairs (ReplacePair): Pairs of strings to replace.

        Returns:
            Callable[[Path], bool]: Function that will receive a file and returns whether it was updated.
        """
        def replace_text(file_path: Path) -> bool:
            """Replace text in a file."""
            updated_text = original_text = file_path.read_text()
            for pair in pairs:
                updated_text = updated_text.replace(pair.Original, pair.Replacement)
            if original_text != updated_text:
                file_path.write_text(updated_text)
                return True
            return False

        return replace_text


def _onerror(func, path, exc_info):
    """Error Handler for shutil rmtree."""
    if not os.access(path, os.W_OK):
        os.chmod(path, stat.S_IWUSR)
        func(path)
    else:
        raise


def run_cmd(cmd, cwd: Path = None) -> int:
    """Run the command and returns the result."""
    logger.print(cmd)
    result = subprocess.run(
        cmd,
        cwd=cwd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding=sys.stdout.encoding,
        errors="ignore",
    )
    if result.returncode != 0:
        logger.log_error(f"Failed with error {result.stdout}")
    else:
        logger.print(f"Successfully executed! Output: \n{result.stdout}")
    return result.returncode


def download_git_model(model_uri: str, model_dir: Path) -> bool:
    """Download model files from GIT repository.

    Args:
        model_uri (str): git clonable uri of a public repo
        model_dir (Path): local directory to clone model to

    Returns:
        bool: True if successful, False otherwise.
    """
    git = shutil.which("git")
    clone_cmd = [git, "clone", model_uri, model_dir]
    result = run_cmd(clone_cmd)
    if result != 0:
        return False
    git_path = model_dir / ".git"
    shutil.rmtree(git_path, onerror=_onerror)
    return True


def run_azcopy(src_uri: str, dstn_uri: str, include_paths: List[str] = None, exclude_paths: List[str] = None,
               as_subdir: bool = True) -> int:
    """Copy blobs between Azure storage accounts or to/from a local dir.

    Args:
        src_uri (str): The source URI.
        dstn_uri (str): The destination URI.
        include_paths (List[str], optional): List of paths to include.
        exclude_paths (List[str], optional): List of paths to exclude.
        as_subdir (bool, optional): If True, copy to a subdirectory under the destination URI.

    Returns:
        int: Return code of azcopy command.
    """
    azcopy = shutil.which("azcopy")
    download_cmd = [azcopy, "copy", src_uri, dstn_uri, "--recursive", "--skip-version-check",
                    "--output-level", "essential"]

    # AzureCloud, USGov, and China clouds should all the trusted Microsoft
    # suffixes built into azcop by default. If the cloud is not one of these,
    # then we need to add the appropriate cloud-specific suffix ourselves.
    if _get_default_cloud_name() not in [AzureEnvironments.ENV_DEFAULT,
                                         AzureEnvironments.ENV_US_GOVERNMENT,
                                         AzureEnvironments.ENV_CHINA]:
        suffix = _get_storage_endpoint_from_metadata()
        if not suffix.startswith("."):
            suffix = "." + suffix
        download_cmd.extend(["--trusted-microsoft-suffixes", suffix])

    if not as_subdir:
        download_cmd.extend(["--as-subdir=false"])
    if include_paths:
        download_cmd.extend(["--include-path", ';'.join(include_paths)])
    if exclude_paths:
        download_cmd.extend(["--exclude-path", ';'.join(exclude_paths)])

    result = run_cmd(download_cmd)
    logger.print(f"azcopy result: {result}")
    return result


def copy_azure_artifacts(src_uri: str, dstn_uri: str, copy_updater: CopyUpdater = None) -> bool:
    """Copy blobs between Azure storage accounts.

    Args:
        src_uri (str): The source storage account URI.
        dstn_uri (str): The destination storage account URI.
        copy_updater (CopyUpdater): CopyUpdater object to update files during azcopy.

    Returns:
        bool: True if successful, False otherwise
    """
    try:
        # Copy between storage accounts, excluding any files to be updated
        update_paths = copy_updater.files if copy_updater else None
        result = run_azcopy(src_uri, dstn_uri, exclude_paths=update_paths)
        if result:
            logger.log_error(f"Failed to copy model files from {src_uri}")
            return False

        # Return if no files to update
        if not copy_updater:
            return True

        # Download files to a temporary directory, update, and then upload
        with tempfile.TemporaryDirectory() as temp_dir:
            # Download files to update
            result = run_azcopy(src_uri, temp_dir, include_paths=update_paths)
            if result:
                logger.log_error(f"Failed to download model files to update from {src_uri}")
                return False

            # Update files under the first subdir, created by azcopy by --as-subdir's default of True
            subdir = Path(temp_dir).iterdir().__next__()
            _ = copy_updater.update_files(subdir)

            # Upload updated files
            result = run_azcopy(temp_dir, dstn_uri, as_subdir=False)
            if result:
                logger.log_error(f"Failed to upload updated model files to {dstn_uri}")
                return False

        return True
    except Exception as e:
        logger.log_error(e)
        return False



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Replace template tags related to Python packages."""

import argparse
import re
from pathlib import Path
from pip._internal.index.collector import LinkCollector
from pip._internal.index.package_finder import PackageFinder
from pip._internal.models.search_scope import SearchScope
from pip._internal.models.selection_prefs import SelectionPreferences
from pip._internal.network.session import PipSession
from typing import List

from azureml.assets.util import logger

# Handles {{latest-pypi-version}} and {{latest-pypi-version:flags}}
LATEST_PYPI_VERSION = re.compile(r"([^\"'\s]+?)(?:(\[.+\]))?([=~]=)\{\{latest-pypi-version(?::(.+))?\}\}")
LATEST_PYPY_VERSION_FLAGS_PRE = "pre"
PYPI_URL = "https://pypi.org/simple"


def create_package_finder(index_urls: List[str]) -> PackageFinder:
    """Create a pip PackageFinder."""
    try:
        link_collector = LinkCollector(
            session=PipSession(),
            search_scope=SearchScope([], index_urls),
        )
    except TypeError:
        # Handle pip>=22.3
        link_collector = LinkCollector(
            session=PipSession(),
            search_scope=SearchScope([], index_urls, False),
        )
    selection_prefs = SelectionPreferences(
        allow_yanked=True,
        ignore_requires_python=True,
    )
    try:
        return PackageFinder.create(
            link_collector=link_collector,
            selection_prefs=selection_prefs,
        )
    except TypeError:
        # Handle pip>=22.0
        return PackageFinder.create(
            link_collector=link_collector,
            selection_prefs=selection_prefs,
            use_deprecated_html5lib=False
        )


def get_latest_package_version(package: str,
                               package_finder: PackageFinder,
                               include_pre: bool = False) -> str:
    """Get the latest version of a Python package.

    Args:
        package (str): Package name.
        package_finder (PackageFinder): PackageFinder to use.
        include_pre (bool, optional): Include pre-release packages. Defaults to False.

    Returns:
        str: Latest package version.
    """
    for _ in range(5):
        try:
            candidates = package_finder.find_all_candidates(package)
            versions = []
            for c in candidates:
                versions.append(c.version)
            if len(versions) > 0:
                versions.sort(reverse=True)
                if include_pre:
                    return str(versions[0])
                else:
                    for v in versions:
                        if not v.is_prerelease:
                            return str(v)
        except Exception as e:
            logger.log_warning(f"Failed to find candidates for {package}: {e}")
            continue
    return None


def pin_packages(contents: str) -> str:
    """Replace Python package template tags in a string.

    Args:
        contents (str): String (likely multi-line) containing tempate tags.

    Returns:
        str: Updated string.
    """
    # Process pip template tags
    package_finder = create_package_finder([PYPI_URL])
    while True:
        match = LATEST_PYPI_VERSION.search(contents)
        if not match:
            break
        package = match.group(1)
        extras = match.group(2) or ""
        selector = match.group(3)
        flags = match.group(4)
        include_pre = True if flags is not None and LATEST_PYPY_VERSION_FLAGS_PRE in flags else False

        logger.log_debug(f"Looking up latest version of {package}")
        version = get_latest_package_version(package, package_finder, include_pre)
        logger.log_debug(f"Latest version of {package} is {version}")
        contents = contents[:match.start()] + f"{package}{extras}{selector}{version}" + contents[match.end():]

    return contents


def transform_file(input_file: Path, output_file: Path = None):
    """Replace Python package template tags in a file.

    Args:
        input_file (Path): Input file.
        output_file (Path, optional): Output file. If not specific, input file will be updated in place.
    """
    # Read file
    with open(input_file) as f:
        contents = f.read()

    # Transform
    contents = pin_packages(contents)

    # Write to stdout or output_file
    if output_file == "-":
        logger.print(contents)
    else:
        if output_file is None:
            output_file = input_file
        with open(output_file, "w") as f:
            f.write(contents)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", type=Path,
                        help="File containing packages to pin to latest versions", required=True)
    parser.add_argument("-o", "--output", type=Path,
                        help="File to which output will be written. Defaults to the input file.")
    args = parser.parse_args()

    output = args.output or args.input
    transform_file(args.input, output)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Script utility methods."""

import difflib
import filecmp
import re
import shutil
from pathlib import Path
from ruamel.yaml import YAML
from typing import List, Tuple, Union

import azureml.assets as assets
from azureml.assets.util import logger
from azureml.assets.config import ValidationException

RELEASE_SUBDIR = "latest"
EXCLUDE_DIR_PREFIX = "!"


# See https://stackoverflow.com/questions/4187564
def are_dir_trees_equal(dir1: Path, dir2: Path, enable_logging: bool = False, ignore_eol: bool = True) -> bool:
    """Compare two directories recursively based on files names and content.

    Args:
        dir1 (Path): First directory
        dir2 (Path): Second directory
        enable_logging (bool, optional): Enable logging for mismatches
        ignore_eol (bool, optional): Ignore EOL differences when comparing files

    Returns:
        bool: True if the directory trees are the same and there were no errors
            while accessing the directories or files, False otherwise.
    """
    dirs_cmp = filecmp.dircmp(dir1, dir2)
    if dirs_cmp.left_only:
        _log_diff(f"Compared {dir1} and {dir2} and found these only in {dir1}: {dirs_cmp.left_only}", enable_logging)
        return False
    if dirs_cmp.right_only:
        _log_diff(f"Compared {dir1} and {dir2} and found these only in {dir2}: {dirs_cmp.right_only}", enable_logging)
        return False
    if dirs_cmp.funny_files:
        _log_diff(f"Compared {dir1} and {dir2} and couldn't compare these: {dirs_cmp.funny_files}", enable_logging)
        return False
    (_, differences, errors) = filecmp.cmpfiles(dir1, dir2, dirs_cmp.common_files, shallow=False)
    if differences:
        _log_diff(f"Compared {dir1} and {dir2} and found differences in: {differences}", enable_logging)
        non_eol_differences = []
        for file in differences:
            if _are_files_equal_ignore_eol(dir1 / file, dir2 / file):
                _log_diff(f"Ignoring differences for {file} because they're only related to EOLs", enable_logging)
            else:
                _log_file_diff(dir1 / file, dir2 / file, enable_logging)
                non_eol_differences.append(file)
        if non_eol_differences:
            return False
    if errors:
        _log_diff(f"Compared {dir1} and {dir2} and couldn't compare these: {errors}", enable_logging)
        return False
    for common_dir in dirs_cmp.common_dirs:
        new_dir1 = dir1 / common_dir
        new_dir2 = dir2 / common_dir
        if not are_dir_trees_equal(new_dir1, new_dir2, enable_logging):
            return False
    return True


def _log_diff(message: str, enabled: bool):
    if enabled:
        logger.log_warning(message)


def _log_file_diff(file1: Path, file2: Path, enabled: bool):
    if enabled:
        with open(file1, "r") as file1_obj, open(file2, "r") as file2_obj:
            file1_contents = file1_obj.readlines()
            file2_contents = file2_obj.readlines()
            diff = difflib.unified_diff(file1_contents, file2_contents, str(file1), str(file2))
            logger.print("".join(diff))


def _are_files_equal_ignore_eol(file1: Path, file2: Path) -> bool:
    with open(file1, "r") as file1_obj, open(file2, "r") as file2_obj:
        while True:
            line1 = file1_obj.readline()
            line2 = file2_obj.readline()
            line1 = line1.rstrip("\n\r") if line1 else None
            line2 = line2.rstrip("\n\r") if line2 else None
            if line1 != line2:
                return False
            if line1 is None and line2 is None:
                return True


def copy_replace_dir(source: Path, dest: Path, paths: List[Path] = None):
    """Copy a directory tree, replacing any existing one.

    Args:
        source (Path): Source directory
        dest (Path): Destination directory
        paths (List[Paths], optional): Specific paths to copy, relative to the source directory
    """
    # Delete destination directory
    if dest.exists():
        shutil.rmtree(dest)

    # Copy source to destination directory
    if not paths:
        # Easy, copy everything
        shutil.copytree(source, dest)
    else:
        # Copy only selected paths
        for path in paths:
            source_path = source / path
            dest_path = dest / path
            if source_path.is_dir():
                Path.mkdir(dest_path, parents=True, exist_ok=True)
            else:
                Path.mkdir(dest_path.parent, parents=True, exist_ok=True)
                shutil.copyfile(source_path, dest_path)


def get_asset_output_dir(asset_config: assets.AssetConfig, output_directory_root: Path,
                         use_version_dir: bool = False) -> Path:
    """Generate the output directory for a given asset.

    Args:
        asset_config (assets.AssetConfig): Asset config
        output_directory_root (Path): Output directory root
        use_version_dir (bool, optional): Add version-specific directory

    Returns:
        Path: The output directory
    """
    version = asset_config.version if use_version_dir else None
    return get_asset_output_dir_from_parts(asset_config.type, asset_config.name, output_directory_root, version)


def get_asset_output_dir_from_parts(type: assets.AssetType, name: str, output_directory_root: Path,
                                    version: str = None) -> Path:
    """Generate the output directory for a given asset.

    Args:
        type (assets.AssetConfig): Asset type
        name (str): Asset name
        output_directory_root (Path): Output directory root
        version (str, optional): Asset version

    Returns:
        Path: The output directory
    """
    output_directory = Path(output_directory_root, type.value, name)
    if version is not None:
        output_directory = output_directory / version

    return output_directory


def get_asset_release_dir(asset_config: assets.AssetConfig, release_directory_root: Path) -> Path:
    """Generate the release directory for a given asset.

    Args:
        asset_config (assets.AssetConfig): Asset config
        release_directory_root (Path): Release directory root

    Returns:
        Path: The release directory
    """
    return get_asset_output_dir(asset_config, release_directory_root / RELEASE_SUBDIR)


def get_asset_release_dir_from_parts(type: assets.AssetType, name: str, release_directory_root: Path) -> Path:
    """Generate the release directory for a given asset.

    Args:
        type (assets.AssetConfig): Asset type
        name (str): Asset name
        release_directory_root (Path): Release directory root

    Returns:
        Path: The release directory
    """
    return get_asset_output_dir_from_parts(type, name, release_directory_root / RELEASE_SUBDIR)


def copy_asset_to_output_dir(asset_config: assets.AssetConfig, output_directory: Path, add_subdir: bool = False,
                             use_version_dir: bool = False) -> Path:
    """Copy asset directory to output directory.

    Args:
        asset_config (assets.AssetConfig): Asset config to copy
        output_directory_root (Path): Output directory root
        add_subdir (bool, optional): Add asset-specific subdirectories to output_directory
        use_version_dir (bool, optional): Store asset in version-specific directory

    Returns:
        Path: The asset's output directory
    """
    if add_subdir:
        output_directory = get_asset_output_dir(asset_config, output_directory, use_version_dir)
    elif use_version_dir:
        output_directory = output_directory / asset_config.version

    common_dir, relative_release_paths = find_common_directory(asset_config.release_paths)
    copy_replace_dir(source=common_dir, dest=output_directory, paths=relative_release_paths)
    return output_directory


def apply_tag_template(full_image_name: str, template: str = None) -> str:
    """Apply a template to an image's tag.

    Args:
        full_image_name (str): The full image name, which must include a tag suffix.
        template (str): Template to use. If desired, use {tag} as a placeholder for the existing tag value.

    Returns:
        str: The transformed image name.
    """
    if template is None:
        return full_image_name

    components = full_image_name.rsplit(":", 1)
    assert len(components) == 2
    components[-1] = template.format(tag=components[-1])
    return ":".join(components)


def apply_version_template(version: str, template: str = None) -> str:
    """Apply a template to a version string.

    Args:
        version (str): The version.
        template (str): Template to use. If desired, use {version} as a placeholder for the existing version string.

    Returns:
        str: The transformed version.
    """
    if template is None:
        return version
    return template.format(version=version)


def _convert_excludes(input_dirs: Union[List[Path], Path],
                      exclude_dirs: List[Path] = None) -> Tuple[List[Path], List[Path]]:
    """Extract directories to exclude from input_dirs and add them to exclude_dirs.

    Args:
        input_dirs (Union[List[Path], Path]): Directories to search in.
        exclude_dirs (List[Path], optional): Directories that should be excluded from the search.

    Returns:
        Tuple[List[Path], List[Path]]: _description_
    """
    if type(input_dirs) is not list:
        input_dirs = [input_dirs]
    if exclude_dirs is not None:
        if type(exclude_dirs) is not list:
            exclude_dirs = [exclude_dirs]
    else:
        exclude_dirs = []

    # Exclude any dirs that start with EXCLUDE_DIR_PREFIX
    new_input_dirs = []
    new_exclude_dirs = []
    for input_dir in input_dirs:
        input_dir_str = str(input_dir)
        if input_dir_str.startswith(EXCLUDE_DIR_PREFIX):
            new_exclude_dirs.append(Path(input_dir_str[len(EXCLUDE_DIR_PREFIX):]))
        else:
            new_input_dirs.append(input_dir)
    if new_exclude_dirs:
        input_dirs = new_input_dirs
        if exclude_dirs:
            exclude_dirs.extend(new_exclude_dirs)
        else:
            exclude_dirs = new_exclude_dirs

    return input_dirs, exclude_dirs


def find_assets(input_dirs: Union[List[Path], Path],
                asset_config_filename: str = assets.DEFAULT_ASSET_FILENAME,
                types: Union[List[assets.AssetType], assets.AssetType] = None,
                changed_files: List[Path] = None,
                exclude_dirs: List[Path] = None,
                pattern: re.Pattern = None) -> List[assets.AssetConfig]:
    """Search directories for assets.

    Args:
        input_dirs (Union[List[Path], Path]): Directories to search in.
        asset_config_filename (str, optional): Asset config filename to search for.
        types (Union[List[assets.AssetType], assets.AssetType], optional): AssetTypes to search for.
        changed_files (List[Path], optional): Changed files, used to filter assets in input_dirs.
        exclude_dirs (Union[List[Path], Path], optional): Directories that should be excluded from the search.
        pattern (re.Pattern, optional): Regex pattern used to filter assets. Defaults to None.

    Returns:
        List[assets.AssetConfig]: Assets found.
    """
    if types is not None and type(types) is not list:
        types = [types]

    # Find and filter assets
    found_assets = []
    for asset_config_file in find_asset_config_files(input_dirs, asset_config_filename, changed_files, exclude_dirs):
        asset_config = assets.AssetConfig(asset_config_file)

        # If specified, skip types not included in filter
        if types and asset_config.type not in types:
            continue

        # If specified, skip assets that don't match the pattern
        if pattern is not None and not pattern.fullmatch(asset_config.full_name):
            continue

        found_assets.append(asset_config)
    return found_assets


def find_asset_config_files(input_dirs: Union[List[Path], Path],
                            asset_config_filename: str,
                            changed_files: List[Path] = None,
                            exclude_dirs: List[Path] = None) -> List[Path]:
    """Search directories for asset config files.

    Args:
        input_dirs (Union[List[Path], Path]): Directories to search in.
        asset_config_filename (str): Asset config filename to search for.
        changed_files (List[Path], optional): Changed files, used to filter assets in input_dirs.
        exclude_dirs (Union[List[Path], Path], optional): Directories that should be excluded from the search.

    Returns:
        List[Path]: Asset config files found.
    """
    input_dirs, exclude_dirs = _convert_excludes(input_dirs, exclude_dirs)
    changed_files_resolved = set([file.resolve() for file in changed_files] if changed_files else [])

    found_assets = []
    for input_dir in input_dirs:
        for file in input_dir.rglob(asset_config_filename):
            # If specified, skip assets when no change in asset, source_code and test_code
            try:
                asset_config = assets.AssetConfig(file)
                test_dir_path = asset_config.pytest_tests_dir_with_path
                test_dir_path = test_dir_path.resolve() if test_dir_path else Path()

                release_paths_resolved = [path.resolve() for path in asset_config.release_paths]
                asset_changed_files = changed_files_resolved & set(release_paths_resolved)
            except ValidationException:
                test_dir_path = Path()
                asset_changed_files = set()

            if changed_files and not asset_changed_files and not any(
                file.parent in f.parents or
                test_dir_path in f.resolve().parents
                for f in changed_files
            ):
                continue

            # If specified, skip excluded directories
            if exclude_dirs and any([d for d in exclude_dirs if d in file.parents]):
                continue

            found_assets.append(file)
    return found_assets


def find_files(input_dirs: Union[List[Path], Path],
               filename: str) -> List[Path]:
    """Search directories for files.

    Args:
        input_dirs (Union[List[Path], Path]): Directories to search in.
        filename (str): Filename to search for.

    Returns:
        List[Path]: Files found (excludes directories).
    """
    found_files = []
    for input_dir in input_dirs:
        for file in input_dir.rglob(filename):
            if file.is_file():
                found_files.append(file)

    return found_files


def find_common_directory(paths: List[Path]) -> Tuple[Path, List[Path]]:
    """Find lowest common directory for a list of Paths.

    Args:
        paths (List[Path]): Paths to consider.

    Returns:
        Tuple[Path, List[Path]]: Common directory and updated Paths which are relative to it.
    """
    lowest_common_dirs = None
    paths_resolved = [p.resolve() for p in paths]
    for path in paths_resolved:
        # Create list of dirs
        path_resolved_parents = list(path.parents)
        if path.is_dir():
            dirs = [path]
            dirs.extend(path_resolved_parents)
        else:
            dirs = path_resolved_parents
        dirs.reverse()

        if lowest_common_dirs is None:
            # Starting point
            lowest_common_dirs = dirs
        else:
            # Find and store common directory
            min_len = min([len(lowest_common_dirs), len(dirs)])
            for i in range(min_len - 1, 0, -1):
                if lowest_common_dirs[i].samefile(dirs[i]):
                    lowest_common_dirs = lowest_common_dirs[0:i + 1]
                    break
            else:
                raise Exception(f"Unable to find a common path between {lowest_common_dirs[-1]} and {dirs[-1]}")

    # Assemble output
    lowest_common_dir = lowest_common_dirs[-1]
    relative_paths = [p.relative_to(lowest_common_dir) for p in paths_resolved]

    return lowest_common_dir, relative_paths


def load_yaml(file_path: str) -> dict:
    """Load yaml utility.

    Args:
        file_path (str): yaml file path

    Returns:
        dict: loaded yaml as dict
    """
    with open(file_path, "r") as f:
        yaml_dict = YAML().load(f)
    return yaml_dict


def dump_yaml(yaml_dict: dict, file_path: str):
    """Dump yaml utility.

    Args:
        yaml_dict (str): dictionary object to dump into yaml.
        file_path (str): File path to store dump result to.
    """
    with open(file_path, "w") as f:
        yaml_dict = YAML().dump(yaml_dict, f)


def retry(times):
    """Retry Decorator.

    Args:
        times (int): The number of times to repeat the wrapped function/method
    """

    def decorator(func):
        def newfn(*args, **kwargs):
            attempt = 1
            while attempt <= times:
                try:
                    return func(*args, **kwargs)
                except Exception:
                    attempt += 1
                    ex_msg = "Exception thrown when attempting to run {}, attempt {} of {}".format(
                        func.__name__, attempt, times
                    )
                    logger.log_warning(ex_msg)
                    if attempt == times:
                        logger.log_warning(
                            "Retried {} times when calling {}, now giving up!".format(times, func.__name__)
                        )
                        raise

        return newfn

    return decorator



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Model utils Operations Class."""

import copy
import azureml.assets as assets
from typing import Tuple
from azure.ai.ml import load_model, MLClient
from azure.ai.ml._utils._registry_utils import get_asset_body_for_registry_storage
from azureml.assets.util import logger
from azureml.assets.config import PathType
from azureml.assets.model.download_utils import CopyUpdater, copy_azure_artifacts, download_git_model
from azureml.assets.deployment_config import AssetVersionUpdate


class RegistryUtils:
    """Registry utils."""

    RETRY_COUNT = 3

    def get_registry_data_reference(model_name: str, model_version: str, ml_client: MLClient) -> Tuple[str, str]:
        """Fetch data reference for asset in the registry."""
        registry_name = ml_client.models._registry_name
        asset_id = f"azureml://registries/{registry_name}/models/{model_name}/versions/{model_version}"
        logger.print(f"getting data reference for asset {asset_id}")
        for cnt in range(1, RegistryUtils.RETRY_COUNT + 1):
            try:
                response = RegistryUtils._get_temp_data_ref(model_name, model_version, ml_client)
                blob_uri = response.blob_reference_for_consumption.blob_uri
                sas_uri = response.blob_reference_for_consumption.credential.additional_properties["sasUri"]
                if not blob_uri or not sas_uri:
                    raise Exception("Error fetching blob or SAS URI")
                return blob_uri, sas_uri
            except Exception as e:
                logger.log_error(f"Exception in fetching data reference. Try #{cnt}. Error: {e}")
        else:
            raise Exception(f"Unable to fetch data reference for asset {asset_id}")

    def _get_temp_data_ref(model_name: str, model_version: str, ml_client: MLClient):
        service_client = ml_client.models._service_client
        registry_name = ml_client.models._registry_name
        body = get_asset_body_for_registry_storage(registry_name, "models", model_name, model_version)
        response = service_client.temporary_data_references.create_or_get_temporary_data_reference(
            name=model_name,
            version=model_version,
            resource_group_name=ml_client.models._resource_group_name,
            registry_name=registry_name,
            body=body,
        )
        return response


class ModelAsset:
    """Asset class for model."""

    def __init__(self, spec_path, model_config, registry_name, temp_dir, copy_updater: CopyUpdater = None):
        """Initialize model asset."""
        self._spec_path = spec_path
        self._model_config = model_config
        self._registry_name = registry_name
        self._temp_dir = temp_dir
        self._copy_updater = copy_updater

        try:
            self._model = load_model(spec_path)
            self._model.description = model_config.description
            self._model.type = model_config.type.value
        except Exception as e:
            logger.error(f"Error in loading model spec file at {spec_path}: {e}")
            return False

    def _publish_to_registry(self, ml_client: MLClient, copy_updater: CopyUpdater = None):
        src_uri = self._model_config.path.uri
        if self._model_config.path.type == PathType.GIT:
            # download model locally
            src_uri = self._temp_dir
            logger.print(f"Cloning model files from git {self._model_config.path.uri}")
            success = download_git_model(self._model_config.path.uri, src_uri)
            logger.print("Completed cloning")
            if not success:
                raise Exception(f"Cloning uri {self._model_config.path.uri} failed")

        try:
            # copy to registry blobstorage
            logger.print("get data ref for registry storage upload")
            blob_uri, sas_uri = RegistryUtils.get_registry_data_reference(
                self._model.name, self._model.version, ml_client
            )
            success = copy_azure_artifacts(src_uri, sas_uri, copy_updater)
            if not success:
                raise Exception("blobstorage copy failed.")
            logger.print("Successfully copied model artifacts to registry storage")
            return blob_uri
        except Exception as e:
            raise Exception(f"Error in copying artifacts to registry storage. Error {e}")


class MLFlowModelAsset(ModelAsset):
    """Asset class for MLflow model."""

    MLMODEL_FILE_NAME = "MLmodel"
    MLFLOW_MODEL_PATH = "mlflow_model_folder"

    def __init__(self, spec_path, model_config, registry_name, temp_dir, copy_updater: CopyUpdater = None):
        """Initialize Mlflow model asset."""
        super().__init__(spec_path, model_config, registry_name, temp_dir, copy_updater)

    def prepare_model(self, ml_client: MLClient):
        """Prepare model for publish."""
        model_registry_path = self._publish_to_registry(ml_client, self._copy_updater)
        self._model.path = model_registry_path + "/" + MLFlowModelAsset.MLFLOW_MODEL_PATH
        return self._model


class CustomModelAsset(ModelAsset):
    """Asset class for custom model."""

    def __init__(self, spec_path, model_config, registry_name, temp_dir, copy_updater: CopyUpdater = None):
        """Initialize custom model asset."""
        super().__init__(spec_path, model_config, registry_name, temp_dir, copy_updater)

    def prepare_model(self, ml_client: MLClient):
        """Prepare model for publish."""
        model_registry_path = self._publish_to_registry(ml_client, self._copy_updater)
        self._model.path = model_registry_path
        return self._model


def prepare_model(spec_path, model_config, temp_dir, ml_client: MLClient, copy_updater: CopyUpdater = None):
    """Prepare model for publish."""
    try:
        logger.print(f"Model type: {model_config.type}")
        registry_name = ml_client.models._registry_name
        if model_config.type == assets.ModelType.CUSTOM:
            model_asset = CustomModelAsset(spec_path, model_config, registry_name, temp_dir, copy_updater)
        elif model_config.type == assets.ModelType.MLFLOW:
            model_asset = MLFlowModelAsset(spec_path, model_config, registry_name, temp_dir, copy_updater)
        else:
            logger.log_error(f"Model type {model_config.type.value} not supported")
            return False
        model = model_asset.prepare_model(ml_client)
        return model, True
    except Exception as e:
        logger.log_error(f"prepare model failed for {spec_path}. Error {e}")
        return None, False


def update_model_metadata(
    model_name: str,
    model_version: str,
    update: AssetVersionUpdate,
    ml_client: MLClient,
    allow_no_op_update: bool = False,
):
    """Update the mutable metadata of already registered Model."""
    try:
        model = ml_client.models.get(name=model_name, version=model_version)
        need_update = False

        # Update tags
        if update.tags:
            updated_tags = copy.deepcopy(model.tags)

            if update.tags.replace is not None:
                updated_tags = update.tags.replace
            else:
                if update.tags.add is not None:
                    for k, v in update.tags.add.items():
                        updated_tags[k] = v
                if update.tags.delete is not None:
                    for k in update.tags.delete:
                        updated_tags.pop(k, None)

            if updated_tags != model.tags:
                logger.print("tags has been updated.")
                model.tags = updated_tags
                need_update = True

        # Update properties
        if update.properties:
            updated_properties = copy.deepcopy(model.properties)
            if update.properties.add is not None:
                for k, v in update.properties.add.items():
                    if k in model.properties and model.properties[k] != v:
                        raise Exception(f"Value of property {k} for model {model_name} cannot "
                                        f"be replaced to {v} without increasing the version.")
                    updated_properties[k] = v

            if updated_properties != model.properties:
                logger.print("properties has been updated.")
                model.properties = updated_properties
                need_update = True

        # Update description
        if update.description is not None and model.description != update.description:
            logger.print("description has been updated")
            model.description = update.description
            need_update = True

        if allow_no_op_update or need_update:
            ml_client.models.create_or_update(model)
            logger.print(f"Model metadata updated successfully for {model_name}")
        else:
            logger.print(f"No update found for model {model_name}. Skipping")
    except Exception as e:
        logger.log_error(f"Failed to update metadata for model : {model_name} : {e}")

    # Archive or restore model based on stage
    try:
        if update.stage == "Archived":
            logger.print(f"Archiving model {model_name} version {model_version}")
            ml_client.models.archive(name=model_name, version=model_version)
        elif update.stage == "Active":
            logger.print(f"Restoring model {model_name} version {model_version}")
            ml_client.models.restore(name=model_name, version=model_version)
    except Exception as e:
        logger.log_error(f"Failed to archive or restore model {model_name} version {model_version}: {e}")



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from .build import build_images
from .pin_versions import transform_file
from .pin_image_versions import pin_images
from .pin_package_versions import pin_packages



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Pin image versions in a Dockerfile."""

import argparse
import json
import re
import urllib.parse
from pathlib import Path
from typing import List, Tuple, Union
from http.client import HTTPResponse
from urllib.request import Request, urlopen

from azureml.assets.util import logger

from tenacity import retry, stop_after_attempt, wait_fixed, retry_if_not_exception_type, retry_if_not_exception_message
from urllib.error import HTTPError

LATEST_TAG = "latest"
# Handles registry/image_name:{{latest-image-tag}} and registry/image_name:{{latest-image-tag:regex}}
LATEST_IMAGE_TAG = re.compile(r"([^\"'\s]+):\{\{latest-image-tag(?::(.+))?\}\}")


@retry(stop=stop_after_attempt(3), wait=wait_fixed(5),
       retry=(retry_if_not_exception_type(HTTPError) & retry_if_not_exception_message(match=r".*404.*")))
def _urlopen_with_retries(request: Union[Request, str]) -> HTTPResponse:
    """Execute urlopen with retries.

    Args:
        request (Union[Request, str]): Request to execute, or URL to open.

    Returns:
        HTTPResponse: Response from urlopen.
    """
    return urlopen(request)


def get_manifest(tag: str, hostname: str, repo: str):
    """Retrieve manifest for an image.

    Args:
        tag (str): Tag for the image.
        hostname (str): Hostname of the container registry.
        repo (str): Repository of the image.

    Returns:
        HTTPResponse: Response from _urlopen_with_retries.
    """
    encoded_tag = urllib.parse.quote(tag, safe="")
    request = Request(f"https://{hostname}/v2/{repo}/manifests/{encoded_tag}",
                      method="HEAD",
                      headers={'Accept': "application/vnd.docker.distribution.manifest.v2+json"})

    return _urlopen_with_retries(request)


def _get_latest_tag_or_digest(image: str, tags: List[str]) -> Tuple[str, str]:
    """Get latest tag or digest for an image.

    Args:
        image (str): Full image name, including hostname of the container registry.
        tags (List[str]): List of tags for the image.

    Returns:
        Tuple[str, str]: Latest tag and digest, or (None, digest) if latest is not found.
    """
    (hostname, repo) = image.split("/", 1)

    # Find another tag corresponding to latest
    latest_tag = None
    latest_digest = None
    for tag in tags:
        # Retrieve digest
        try:
            response = get_manifest(tag, hostname, repo)
        except Exception as e:
            raise Exception(f"Failed to retrieve manifest for {repo}:{tag}: {e}")

        digest = response.info()['Docker-Content-Digest']

        if tag == LATEST_TAG:
            # Store latest digest for comparison
            latest_digest = digest
        elif digest == latest_digest:
            # Found matching digest
            latest_tag = tag
            break

    return latest_tag, latest_digest


def _get_latest_image_suffix(image: str, regex: re.Pattern = None) -> str:
    """Get suffix to use for latest image tag or digest.

    Args:
        image (str): Full image name, including hostname of the container registry.
        regex (re.Pattern): Regex to use to filter tags.

    Returns:
        str: Suffix as :tag or @digest.
    """
    (hostname, repo) = image.split("/", 1)

    # Retrieve tags
    try:
        response = _urlopen_with_retries(f"https://{hostname}/v2/{repo}/tags/list")
    except Exception as e:
        raise Exception(f"Failed to retrieve tags for {repo}: {e}")
    tags = json.loads(response.read().decode("utf-8")).get("tags", [])

    # Filter tags and sort in descending order because this should be faster
    tags_sorted = sorted([t for t in tags if t != LATEST_TAG and
                         (regex is None or regex.fullmatch(t) is not None)], reverse=True)

    # Handle regex
    if regex is not None:
        # Use the most recent matching tag
        if tags_sorted:
            latest_tag = tags_sorted[0]
        else:
            raise Exception(f"{image} does not have tag that matches {regex}")
    else:
        # Ensure latest is present
        if LATEST_TAG not in tags:
            raise Exception(f"{image} does not have a {LATEST_TAG} tag")

        # Insert latest at the beginning to ensure we get its digest first
        tags_sorted.insert(0, LATEST_TAG)

        # Find another tag corresponding to latest, or default to digest
        latest_tag, latest_digest = _get_latest_tag_or_digest(image, tags_sorted)

    # Return tag or digest
    if latest_tag is not None:
        return f":{latest_tag}"
    else:
        logger.log_warning(f"Using digest for {image} because a non-{LATEST_TAG} was not found")
        return f"@{latest_digest}"


def pin_images(contents: str) -> str:
    """Pin images in a Dockerfile.

    Args:
        contents (str): Contents of a Dockerfile.

    Returns:
        str: Contents of the Dockerfile with images pinned to latest versions.
    """
    # Process MCR template tags
    while True:
        match = LATEST_IMAGE_TAG.search(contents)
        if not match:
            break
        repo = match.group(1)
        regex = match.group(2)
        message = f"Finding latest image tag/digest for {repo}"
        if regex is not None:
            message += f" matching {regex}"
            regex = re.compile(regex)
        logger.log_debug(message)
        suffix = _get_latest_image_suffix(repo, regex)
        logger.log_debug(f"Latest image reference is {repo}{suffix}")
        contents = contents[:match.start()] + f"{repo}{suffix}" + contents[match.end():]

    return contents


def transform_file(input_file: Path, output_file: Path = None):
    """Transform a file.

    Args:
        input_file (Path): File to transform.
        output_file (Path): File to which output will be written. Defaults to the input file.
    """
    # Read file
    with open(input_file) as f:
        contents = f.read()

    # Transform
    contents = pin_images(contents)

    # Write to stdout or output_file
    if output_file == "-":
        logger.print(contents)
    else:
        if output_file is None:
            output_file = input_file
        with open(output_file, "w") as f:
            f.write(contents)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", type=Path,
                        help="File containing images to pin to latest versions", required=True)
    parser.add_argument("-o", "--output", type=Path,
                        help="File to which output will be written. Defaults to the input file.")
    args = parser.parse_args()

    output = args.output or args.input
    transform_file(args.input, output)
